use nom::combinator::map;
use nom::combinator::not;

use crate::parser::error::CustomError;
use crate::parser::error::MyError;

use super::error::Res;
use super::parser_with_context::parser_with_context;
use super::source::Source;
use super::Context;

#[derive(Debug)]
pub enum Object<'s> {
    TextMarkup(TextMarkup<'s>),
    PlainText(PlainText<'s>),
    RegularLink(RegularLink<'s>),
}

#[derive(Debug)]
pub struct TextMarkup<'s> {
    pub source: &'s str,
}

#[derive(Debug)]
pub struct PlainText<'s> {
    pub source: &'s str,
}

#[derive(Debug)]
pub struct RegularLink<'s> {
    pub source: &'s str,
}

impl<'s> Source<'s> for Object<'s> {
    fn get_source(&'s self) -> &'s str {
        match self {
            Object::TextMarkup(obj) => obj.source,
            Object::PlainText(obj) => obj.source,
            Object::RegularLink(obj) => obj.source,
        }
    }
}

pub fn standard_set_object<'r, 's>(
    context: Context<'r, 's>,
    input: &'s str,
) -> Res<&'s str, Object<'s>> {
    not(|i| context.check_exit_matcher(i))(input)?;

    let plain_text_matcher = parser_with_context!(plain_text)(context);

    map(plain_text_matcher, Object::PlainText)(input)
}

fn plain_text<'r, 's>(context: Context<'r, 's>, input: &'s str) -> Res<&'s str, PlainText<'s>> {
    if input.len() == 0 {
        return Err(nom::Err::Error(CustomError::MyError(MyError(
            "Zero input length to plain_text.",
        ))));
    }
    // not(|i| context.check_exit_matcher(i))(input)?;
    let mut current_input = input.char_indices();
    loop {
        match current_input.next() {
            Some((offset, _char)) => {
                let remaining = &input[offset..];
                let exit_matcher_status = not(|i| context.check_exit_matcher(i))(remaining);
                if exit_matcher_status.is_err() {
                    if offset == 0 {
                        // If we're at the start of the input, then nothing is plain text, so fire an error for zero-length match.
                        exit_matcher_status?;
                    } else {
                        return Ok((
                            &input[offset..],
                            PlainText {
                                source: &input[..offset],
                            },
                        ));
                    }
                }
            }
            None => {
                // We hit the end of the file, so all input must be plain text
                return Ok((&input[input.len()..], PlainText { source: input }));
            }
        };
    }
}

#[cfg(test)]
mod tests {
    use crate::parser::parser_context::ContextElement;
    use crate::parser::parser_context::ContextTree;

    use super::*;

    #[test]
    fn plain_text_simple() {
        let input = "foobarbaz";
        let initial_context: ContextTree<'_, '_> = ContextTree::new();
        let document_context =
            initial_context.with_additional_node(ContextElement::DocumentRoot(input));
        let plain_text_matcher = parser_with_context!(plain_text)(&document_context);
        let (remaining, result) = map(plain_text_matcher, Object::PlainText)(input).unwrap();
        assert_eq!(remaining, "");
        assert_eq!(result.get_source(), input);
    }
}
